import FollowCard from '../follow_card/follow_card.vue'
import Conversation from '../conversation/conversation.vue'
import TabSwitcher from 'src/components/tab_switcher/tab_switcher.jsx'
import SearchFilters from './search_filters.vue'
import map from 'lodash/map'
import { library } from '@fortawesome/fontawesome-svg-core'
import {
  faCircleNotch,
  faSearch
} from '@fortawesome/free-solid-svg-icons'
import { uniqBy } from 'lodash'

library.add(
  faCircleNotch,
  faSearch
)

const allSearchTypes = Object.freeze(['statuses', 'media', 'accounts', 'hashtags'])

const Search = {
  components: {
    FollowCard,
    Conversation,
    TabSwitcher,
    SearchFilters
  },
  props: [
    'query'
  ],
  data () {
    return {
      queryCount: 0,
      loadedInitially: false,
      loading: Object.fromEntries(
        allSearchTypes.map((searchType) => [searchType, false])
      ),
      searchTerm: this.query || '',
      userIds: [],
      statuses: [],
      media: [],
      hashtags: [],
      allTabs: allSearchTypes,
      currentResultTab: 'statuses',
      hasUserSelectedTab: false,

      statusesOffset: 0,
      lastStatusFetchCount: 0,
      mediaOffset: 0,
      lastMediaFetchCount: 0,
      lastQuery: '',
      filter: {}
    }
  },
  computed: {
    users () {
      return this.userIds.map(userId => this.$store.getters.findUser(userId))
    },
    visibleStatuses () {
      const allStatusesObject = this.$store.state.statuses.allStatusesObject

      return this.statuses.filter(status =>
        allStatusesObject[status.id] && !allStatusesObject[status.id].deleted
      )
    },
    visibleMedia () {
      const allStatusesObject = this.$store.state.statuses.allStatusesObject

      return this.media.filter(status =>
        allStatusesObject[status.id] && !allStatusesObject[status.id].deleted
      )
    },
    canSearchMediaPosts () {
      return this.$store.state.instance.searchTypeMediaEnabled === true
    },
    hasAtLeastOneResult () {
      return allSearchTypes
        .some((searchType) => this.getVisibleLength(searchType) > 0)
    }
  },
  mounted () {
    this.search(this.query)
  },
  watch: {
    query (newValue) {
      this.searchTerm = newValue
      this.search(newValue)
    },
    filter: {
      deep: true,
      handler () {
        this.lastQuery = "" // invalidate state
        this.search(this.searchTerm)
      }
    }
  },
  methods: {
    newQuery (query) {
      if (this.lastQuery === query && !this.loading[this.currentResultTab]) {
        // Handle search retries
        this.lastQuery = "" // invalidate state
        this.search(query)
      } else {
        this.$router.push({ name: 'search', query: { query } })
        this.$refs.searchInput.focus()
      }
    },
    async search (query, searchType = null) {
      if (!query) {
        for (const searchType of allSearchTypes) {
          this.loading[searchType] = false
        }
        return
      }

      const localQueryCount = ++this.queryCount

      const isNewSearch = this.lastQuery !== query
      this.$refs.searchInput.blur()
      if (isNewSearch) {
        this.userIds = []
        this.hashtags = []
        this.statuses = []
        this.media = []

        this.statusesOffset = 0
        this.lastStatusFetchCount = 0
        this.mediaOffset = 0
        this.lastMediaFetchCount = 0
      }

      let searchTypes = allSearchTypes
      if (searchType) {
        // Search only for `searchType` if it is provided
        searchTypes = [searchType]
      } else if (this.hasUserSelectedTab && this.currentResultTab !== 'statuses') {
        // Start the search from the tab that the user has selected
        // No need to sort if userPreferredTab === 'statuses'
        searchTypes = [
          this.currentResultTab,
          ...allSearchTypes.filter((tab) => tab !== this.currentResultTab)
        ]
      }

      for (const searchType of searchTypes) {
        this.loading[searchType] = true
      }

      let oldStatusesLength = this.statuses.length
      let oldMediaLength = this.media.length

      let skipMediaSearch = !this.canSearchMediaPosts

      for (const searchType of searchTypes) {
        try {
          if (searchType === 'media' && skipMediaSearch) {
            continue
          }

          let searchOffset
          if (searchType === 'statuses') {
            searchOffset = this.statusesOffset
          } else if (searchType === 'media') {
            searchOffset = this.mediaOffset
          }

          const data = await this.$store.dispatch('search', {
            q: query,
            resolve: true,
            offset: searchOffset,
            'type': searchType,
            following:
              'followingOnly' in this.filter && this.filter.followingOnly,
            local: 'localOnly' in this.filter && this.filter.localOnly
          })

          if (localQueryCount !== this.queryCount) {
            // Query count differs, there should be a newer query
            return
          }

          // Always append to old results. If new results are empty, this doesn't change anything
          this.userIds = this.userIds.concat(map(data.accounts, 'id'))
          this.statuses = uniqBy(this.statuses.concat(data.statuses), 'id')
          if ('media' in data) {
            this.media = uniqBy(this.media.concat(data.media), 'id')
          }
          this.hashtags = this.hashtags.concat(data.hashtags)

          if (isNewSearch) {
            if (!this.hasUserSelectedTab) {
              this.currentResultTab = this.getFirstTabWithResults()
            }
            if (searchType === 'statuses' && data.statuses.length === 0) {
              // Safe to assume that there are no media posts
              skipMediaSearch = true
            }
          }
        } catch (error) {
          console.error(error)
        } finally {
          if (localQueryCount !== this.queryCount) {
            // Skip cleanups if there's a newer query
            return
          }

          if (!this.loadedInitially && this.hasAtLeastOneResult) {
            // Show results on the first meaningful response
            this.loadedInitially = true
          }
          this.loading[searchType] = false

          if (searchType === 'statuses') {
            // Offset from whatever we already have
            this.statusesOffset = this.statuses.length
            // Because the amount of new statuses can actually be zero, compare to old length instead
            this.lastStatusFetchCount = this.statuses.length - oldStatusesLength
          } else if (searchType === 'media') {
            this.mediaOffset = this.media.length
            this.lastMediaFetchCount = this.media.length - oldMediaLength
          }
        }
      }

      this.lastQuery = query
      this.loadedInitially = true
      for (const searchType of allSearchTypes) {
        this.loading[searchType] = false
      }
    },
    resultCount (tab) {
      const length = this.getVisibleLength(tab)

      if (length === 0 || !this.loadedInitially) {
        return ''
      }

      if (
        (tab === 'statuses' && this.lastStatusFetchCount !== 0) ||
        (tab === 'media' && this.lastMediaFetchCount !== 0)
      ) {
        return ` (${length}+)`
      }

      return ` (${length})`
    },
    onResultTabSwitch (key) {
      this.currentResultTab = key
      this.hasUserSelectedTab = true
    },
    getFirstTabWithResults () {
      for (const tab of allSearchTypes) {
        if (this.getVisibleLength(tab) > 0) {
          return tab
        }
      }

      return 'statuses'
    },
    lastHistoryRecord (hashtag) {
      return hashtag.history && hashtag.history[0]
    },
    getVisibleLength (tab) {
      if (tab === 'statuses') {
        return this.visibleStatuses.length
      } else if (tab === 'media') {
        return this.visibleMedia.length
      } else if (tab === 'accounts') {
        return this.users.length
      } else if (tab === 'hashtags') {
        return this.hashtags.length
      }
    }
  }
}

export default Search
